//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "TMainForm.h"
#include "TDebuggingForm.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsFunctionName( TStringList* function_name_strings, AnsiString function_name )
{
for (int i=0;i<function_name_strings->Count;i++) if (function_name_strings->Strings[i]==function_name) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsInterfaceName( TStringList* Interface_name_strings, AnsiString interface_name )
{
for (int i=0;i<Interface_name_strings->Count;i++) if (Interface_name_strings->Strings[i]==interface_name) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::HasFunctionName( TStringList* function_name_strings, AnsiString function_name )
{
for (int i=0;i<function_name_strings->Count;i++) if (function_name_strings->Strings[i].Pos(function_name)!=0) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::HasInterfaceName( TStringList* Interface_name_strings, AnsiString interface_name )
{
for (int i=0;i<Interface_name_strings->Count;i++) if (Interface_name_strings->Strings[i].Pos(interface_name)!=0) return true;
return false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ImportInterfaces( AnsiString pFileName )
{
DebuggingForm->Visible = true;
DebuggingForm->Caption = "Importing Interfaces.... debugging output";

//---------------------------------------------------------------------------
// Create storage for information to be stored in the database.
//---------------------------------------------------------------------------
StoredInterfaceNames = new TStringList();
StoredInterfaceDescriptions = new TStringList();
StoredFunctionNames = new TStringList();
StoredFunctionDescriptions = new TStringList();
StoredFunctionReturnTypes = new TStringList();
StoredFunctionReturnValues = new TStringList();
StoredParameterTypes = new TStringList();
StoredParameterNames = new TStringList();
StoredParameterModifiers = new TStringList();   //can be "*" or "FAR *"
StoredParameterDescriptions = new TStringList();
fFlagNames = new TStringList();
fFlagDescriptions = new TStringList();

try {   // try..catch is for exception handling

//---------------------------------------------------------------------------
// Build a list of the various Interfaces included in this import file
//---------------------------------------------------------------------------
Dbg( "Building a list of the various Interface names in the file: '" + pFileName + "'" );
bool found_first_repeated_Interface = false;
int temp_count = 0;
do
    {
    // check the current row to see if it contains a unique structure name
    if (IsInterfaceName(StoredInterfaceNames, ImportString))
        {
        Dbg( "Found a repetition of the Interface name: '" + ImportString + "'... assuming the Interface List is finished." );
        found_first_repeated_Interface = true;
        break;  // continue at end of this do..while loop
        }
    // if we didn't find a unique Interface name
    if (!found_first_repeated_Interface)
        {
        // check that the row conforms to the Interface naming rules
        if (IsValidInterfaceName(ImportString))
            {
            StoredInterfaceNames->Add( ImportString );
            Dbg( ">>> Interface #" + IntToStr(StoredInterfaceNames->Count) + ": '" + ImportString + "' <<<" );
            ImportRow++;
//            if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");
            }
        else ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");
        }
    }
while (!found_first_repeated_Interface);

//---------------------------------------------------------------------------
// scan through each interface in turn, parsing as we go
//---------------------------------------------------------------------------
// since we now know how many interfaces we need to parse, attempt to parse that many interfaces
for (int interface_index=0;interface_index<StoredInterfaceNames->Count;interface_index++)
    {
    Dbg( ">>> About to Parse Interface #" + IntToStr(interface_index) + " <<<" );
//    if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned but did not find the interface definition for '" + StoredInterfaceNames->Strings[interface_index] + "'" );

    //-----------------------------------------------------------------------
    // create internal storage for this interface
    //-----------------------------------------------------------------------
    TStringList* TempFunctionNames = new TStringList();
	TStringList* TempFunctionDescriptions = new TStringList();
	TStringList* TempFunctionReturnTypes = new TStringList();
	TStringList* TempFunctionReturnValues = new TStringList();
	TStringList* TempParameterNames = new TStringList();
	TStringList* TempParameterDescriptions = new TStringList();
	TStringList* TempParameterTypes = new TStringList();
	TStringList* TempParameterModifiers = new TStringList();
    TStringList* TempFlagNames = new TStringList();
	TStringList* TempFlagDescriptions = new TStringList();

    //-----------------------------------------------------------------------
	// check that an appropriate Interface name is on this row
    //-----------------------------------------------------------------------
    if (ImportString!=StoredInterfaceNames->Strings[interface_index]) ERROR( "Did not find the expected Interface name: '" + StoredInterfaceNames->Strings[interface_index] + "'" );
    Dbg( ">>> Parsing Interface #" + IntToStr(interface_index) + ": '" + StoredInterfaceNames->Strings[interface_index] + "' <<<" );
    ImportRow++;
//    if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before after was repeated (No Interface Definition?)");

    //-----------------------------------------------------------------------
    // the next row is assumed to contain the Interface description
    //-----------------------------------------------------------------------
    StoredInterfaceDescriptions->Add( ImportString );
    Dbg( "Adding Interface #" + IntToStr(interface_index) + " Description: '" + ImportString + "'" );
    ImportRow++;
//    if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");

    //-----------------------------------------------------------------------
    // build a list of the Functions associated with this interface
    //-----------------------------------------------------------------------
    Dbg( "Building list of Function Names for Interface #" + IntToStr(interface_index) );
    // the row following the function names always (1) contains a "::"
    // or (2) is not a single word (token)
    bool end_of_function_names = false;
    do
        {
        if ((ImportString.Pos("::") > 0) || !IsSingleToken(ImportString))
            {
            Dbg( "Found the end of the function_names list." );
            end_of_function_names = true;
            continue;
            }
   		TempFunctionNames->Add( ImportString );
        Dbg( "Adding Function #" + IntToStr(TempFunctionNames->Count) + " Name: '" + ImportString + "'" );
        ImportRow++;
//        if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");
        }
    while (!end_of_function_names);

    //-----------------------------------------------------------------------
    // skip any miscellaneous declarations (such as typedef's) till we
    // find a function declaration
    //-----------------------------------------------------------------------
    while ((ImportString.Pos("::") == 0) || !IsSingleToken(ImportString))
        {
        ImportRow++ ;
//      if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");
        }

    //-----------------------------------------------------------------------
    // process the functions for the current Interface
    //-----------------------------------------------------------------------
    for (int function_index=0;function_index<TempFunctionNames->Count;function_index++)
    	{
        Dbg(">>> Expecting to find a function declaration for '" + TempFunctionNames->Strings[function_index] + "'" );
//        if (ImportRow >= ImportStrings->Count) ERROR( "scanned all rows but did not find the relevant function definition for '" + TempFunctionNames->Strings[function_index] + "'" );

        // "IDirect3DRMArray::GetSize"
        if (TempFunctionNames->Strings[function_index] == "AddRef")
            {
            Dbg("Special Case: Ignoring AddRef() function");
            continue;
            }
        if (TempFunctionNames->Strings[function_index] == "QueryInterface")
            {
            Dbg("Special Case: Ignoring QueryInterface() function");
            continue;
            }
        if (TempFunctionNames->Strings[function_index] == "Release")
            {
            Dbg("Special Case: Ignoring Release() function");
            continue;
            }

        //-------------------------------------------------------------------
        // process the function description
        //-------------------------------------------------------------------
	    if (ImportString.Pos("::")==0) ERROR( "Did not find the '::' keyword as expected." );
// 	   	Dbg( "Row contained a valid '::' keyword. Assuming Function Description follows" );
 	   	ImportRow++;
//        if (ImportRow >= ImportStrings->Count) ERROR( "All rows were scanned before an Interface Name was repeated (No Interface Definition?)");

		// now we have a description for the Function
		if (IsSingleToken(ImportString))
        	{
        	//error, not a description
        	ERROR( "Did not find Function Description as expected." );
		   	//ImportRow++;
        	}
    	else
    		{
       	 	TempFunctionDescriptions->Add( ImportString );
            Dbg( "Adding Function #" + IntToStr(TempFunctionDescriptions->Count) + " Description: '" + ImportString + "'" );
 	   		ImportRow++;
       		}

        //-------------------------------------------------------------------
        // process the function declaration, building list of various
        // parameters for the function
        //-------------------------------------------------------------------
        // next ... is DWORD GetSize( );
        // we get the "DWORD" as FunctionReturnType
        if (ImportString.Pos("(") == 0) ERROR ("Did not find Function Return Type line as expected");
        TempFunctionReturnTypes->Add(ImportString.SubString(1,ImportString.Pos(" ")));
        Dbg( "Adding Function #" + IntToStr(TempFunctionDescriptions->Count) + " Return Type: '" + ImportString.SubString(1,ImportString.Pos(" ")) + "'" );
        bool parameter_list_found = false;

        TempParameterNames->Clear();
        TempParameterModifiers->Clear();
        TempParameterTypes->Clear();
        do
        	{
        	if (ImportString.Pos(");") > 0) continue;  // and expect return value
     	   	ImportRow++;
            parameter_list_found = true;
         	if (ImportString.Pos(");") > 0) continue;  // and expect return value
	        if (ImportString.Pos("*") != 0)
            	// found "*", ParameterName is ltrim of the rest of the substring
                // this takes the FAR into account when we create the parameter modifier
                // which can be "*" or "FAR *"
            	{
                int asterisk = ImportString.Pos("*");
                int comma = ImportString.Pos(",")-1;
                if (comma==-1) comma = ImportString.Length();
                int space = ImportString.Pos(" ");

                AnsiString param_name = ImportString.SubString(asterisk+1,comma-asterisk).TrimLeft();
                AnsiString param_modifiers = ImportString.SubString(space+1,asterisk-space);
                AnsiString param_type = ImportString.SubString(1,space-1);
                TempParameterNames->Add( param_name );
                TempParameterModifiers->Add( param_modifiers );
                TempParameterTypes->Add( param_type );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Name : '" + param_name + "'" );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Type : '" + param_type + "'" );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Mods : '" + param_modifiers + "'" );
                }
            else
            	{
                int comma = ImportString.Pos(",")-1;
                if (comma==-1) comma = ImportString.Length();
                int space = ImportString.Pos(" ");
                AnsiString param_name = ImportString.SubString(space+1,comma-space).TrimLeft();
                AnsiString param_type = ImportString.SubString(1,space-1);
                TempParameterNames->Add( param_name ); //some trimming required
                TempParameterTypes->Add( param_type );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Name : '" + param_name + "'" );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Type : '" + param_type + "'" );
                Dbg( "Adding Param #" + IntToStr(TempParameterNames->Count) + " Mods : NONE");
                }
            }
        while (ImportString.Pos(");") == 0);
        ImportRow++;

        //-------------------------------------------------------------------
        // process the parameter descriptions....
        //-------------------------------------------------------------------
      	// we can now parse the descriptions of the various parameters of the interface.
     	// note that some parameters also contain descriptions of various flags.
		if (parameter_list_found)
        	{
       		if (ImportString=="Parameters") ImportRow++; // skip the parameters line
   			bool checking_last_parameter_flags = false;
   			for (int parameter_index=0;parameter_index<TempParameterNames->Count;parameter_index++)
   		 	    {
//                if (ImportRow >= ImportStrings->Count) ERROR( "scanned all rows but did not find the relevant parameter definition for '" + TempParameterNames->Strings[parameter_index] + "'" );
   		    	// check for the premature end of the parameter descriptions
   		    	if (IsKeyword(ImportString,"QuickInfo"))
   	       		 	{
   		    	    // special case: if we're checking for flags for the last member, and the check failed
   	        		// do not check for "this and that" rules.
   	        		if (!checking_last_parameter_flags) ERROR( "found a 'QuickInfo' section before all parameter descriptions were defined." );
   	        		Dbg( "found a 'QuickInfo' section after checking all parameter descriptions. Assuming parameter descriptions are completed" );
   	        		break;
   	        		}
    	    	if (IsKeyword(ImportString,"Remarks"))
    	    	    {
    	    	    // special case: if we're checking for flags for the last member, and the check failed
    	    	    // do not check for "this and that" rules.
    	    	    if (!checking_last_parameter_flags) ERROR( "found a 'Remarks' section before all parameter descriptions were defined." );
    	    	    Dbg( "found a 'Remarks' section after checking all parameter descriptions. Assuming parameter descriptions are completed" );
    	    	    break;
    	    	    }
                // if the current line contains the name of the appropriate member, ignore the member name &
    	    	// store the description that follows immediately after this row.
    	    	if (ImportString==TempParameterNames->Strings[parameter_index])
    	    	    {
//    	    	    Msg( "Parsing Interface Parameter: '" + StoredInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
    	        	ImportRow++;

    	        	TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                    Dbg( "Adding Param #" + IntToStr(parameter_index) + " Description : '" + ImportString + "'" );
    	        	ImportRow++;

    	        	// check if there are any flags following the member description
   	         		Dbg( "Checking to see if any flags are declared following the parameter description" );
                	if (IsFlag(ImportString))
             		   	{
             		   	Dbg( "Found a Flag immediately following the parameter description." );
             		   	do
             		  		{
//                	   		Dbg( "adding a parameter flag: '" + ImportString + "'" );
//             		       	Msg( "Found Interface Parameter Flag: '" + StoredInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" + "::" + ImportString );
             		       	TempFlagNames->Add( ImportString);
                            Dbg( "Adding Flag Name: '" + ImportString + "'" );
             		       	// check for flag descriptions declared on the same line as the flag name
                	       	int temp_pos = ImportString.Pos(" ");
              		 	   	if (temp_pos!=0)
                	        	{
//                	        	Dbg( "the parameter flag row also contains the relevant flag description" );
                	  	 		AnsiString flag_description = ImportString.SubString(temp_pos,ImportString.Length()-temp_pos).TrimLeft();
//                	        	Dbg( "adding the parameter flag description: '" + flag_description + "'" );
                	        	TempFlagDescriptions->Add(EncodeValue("FlagInterfaceIndex",interface_index) + EncodeValue("FlagParameterIndex",parameter_index) + flag_description);
                                Dbg( "Adding Flag Description : '" + ImportString + "'" );
                	        	}
                	       	else
                	        	// check just to make sure that the flags do not have descriptions with them...
                	        	if (IsFlag(ImportString))
                	            	{
//                	            	Dbg( "the parameter flag description is assumed to be on the next row..." );
                	            	ImportRow++;
//                	            	Dbg("adding a parameter flag description: '" + ImportString + "'" );
                	            	TempFlagDescriptions->Add(EncodeValue("FlagInterfaceIndex",interface_index) + EncodeValue("FlagParameterIndex",parameter_index) + ImportString);
                                    Dbg( "Adding Flag Description : '" + ImportString + "'" );
                	            	ImportRow++;
                	            	}
                	        	else
                	            	ERROR( "a flag was encountered followed immediately by another flag with no descriptions !" );
                	        }
                	    while (IsFlag(ImportString));
                		}
            		else
                		Dbg( "there are no flags following the parameter description" );
            		}
        		else
                	// it's not a simple parameter, where one was expected.... maybe it's a case of "this and that".... or "this, that, and something else"
            		{
//            		Dbg( "did not find a parameter name where expected... checking 'this and that' rules..." );
            		// if there is another parameter name following this one....
            		if (parameter_index<TempParameterNames->Count-1) // room for this and that
                		{
//                		Dbg( "...checking 'this and that' rules");
                		AnsiString first_parameter_name;
                		AnsiString second_parameter_name;
                		if (IsThisAndThat(TempParameterNames,ImportString,&first_parameter_name,&second_parameter_name))
                		    {
             		   	    Dbg( "Found a 'This and That' Flag immediately following the parameter description." );
//                		    Dbg( "the row fits the rules for a 'this and that' type description" );
                		    ImportRow++;    // skip past the "this and that" row
//                		    Dbg( "assuming row #" + IntToStr(ImportRow) + " is the description for parameters: '" + first_parameter_name + "' and '" + second_parameter_name + "'" );
                		    TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                		    TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                            Dbg( "Adding Flag Description : '" + ImportString + "'" );
                            Dbg( "Adding Flag Description : '" + ImportString + "'" );
                		    ImportRow++;
                		    parameter_index++;
                		    }
                		else
                		    {
//                		    Dbg( "the row did not fit the 'this and that' rule. Checking 'this, that, and something else' rule..." );
                		    TStringList* temp_parameter_names = new TStringList();
                		    if (IsThisAndThatAndSomethingElse(TempParameterNames,ImportString, temp_parameter_names ))
                		        {
             		   	        Dbg( "Found a 'This, That and SomethingElse' Flag immediately following the parameter description." );
                		        ImportRow++;    // skip past the "this, that and something else" row
                		        for (int i=0;i<temp_parameter_names->Count;i++)
                		            {
//                		            Dbg( "adding parameter description for parameter: '" + temp_parameter_names->Strings[i] + "'" );
                		            TempParameterDescriptions->Add( EncodeValue("InterfaceIndex",interface_index) + ImportString );
                                    Dbg( "Adding Flag Description : '" + ImportString + "'" );
                		            }
                		        if (temp_parameter_names->Count>0) parameter_index += temp_parameter_names->Count-1;
                		        ImportRow++;
                		        }
               			    else
                		        {
                	            Dbg( "Found the following parameter names when checking 'this, that, and something else' rule" );
                	        	for (int i=0;i<temp_parameter_names->Count;i++) Dbg( "parameter #" + IntToStr(i) + "=='" + temp_parameter_names->Strings[i] + "'" );
             		 			temp_parameter_names->Clear();
             					delete temp_parameter_names;
             		 			ERROR( "Expected but did not find a particular parameter name & description: '" + StoredInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
								}
 							temp_parameter_names->Clear();
           	    	     	delete temp_parameter_names;
                	    	}
                		}
            		else
            	    	ERROR( "Expected but did not find a particular parameter name & description: '" + StoredInterfaceNames->Strings[interface_index] + "::" + TempParameterNames->Strings[parameter_index] + "'" );
            	    }   // end of "this & that" rules

//                if (ImportString=="Return Values")
//                    {
//                   	Dbg("found 'Return Values' #1");
//                   	ImportRow ++ ;
//                   	// read a string into database
//                   	TempFunctionReturnValues->Add(ImportString);
//                   	Dbg("'Function Return Values' are" + ImportString);
//                   	ImportRow ++ ;
//                   	}
        		} // end of "for (parameter_index=0;)"
            } // end of "if parameter list found"

    	// we now expect either a "Remarks", "Return values", new Interface or a "QuickInfo" section
    	// check just to make sure
    	// if it is a 'Remarks' section, simply skip it till we find a "QuickInfo" section.

        if (ImportString=="Return Values")
            {
            Dbg("found the 'Return Values' keyword... skipping section entirely");
            ImportRow ++ ;
            // read a string into database
            TempFunctionReturnValues->Add(ImportString);
            Dbg("'Function Return Values' are: '" + ImportString + "'");
            ImportRow++;
            }
   	    while ((ImportString!="QuickInfo") && (ImportString!=""))
   	        {
   	        ImportRow++;
            if ( (ImportString.Pos("::")!=0) && HasInterfaceName(StoredInterfaceNames,ImportString) &&
                HasFunctionName(TempFunctionNames,ImportString)) 
                {
                Dbg("Found a new declaration before the 'QuickInfo' section. Ignoring 'QuickInfo' section");
                break;
                }
//   	        if (ImportRow>=ImportStrings->Count) ERROR("did not find the QuickInfo section at the end of a function");
            }
        if (ImportString=="QuickInfo")
            {
            Dbg("Found a 'QuickInfo' section... skipping it entirely");
            while ((ImportString.Pos("::")==0) && !IsInterfaceName(StoredInterfaceNames,ImportString) && (ImportString!=""))
                {
                ImportRow++;
    //   	        if (ImportRow>=ImportStrings->Count)
    //                if (interface_index==StoredInterfaceNames->Count-1) break;
    //                else ERROR("did not find the QuickInfo section at the end of a function");
                }
            }
        }
    // yey. the structure definition has been successfully parsed.
    Dbg( ">>>>>> Definition of Interface '" + StoredInterfaceNames->Strings[interface_index] + "' was completed successfully! <<<<<<" );

    // copy the stuff out of temporary storage into more permanent storage, then clear the temporary storage
/*
    int i;
    for (i=0;i<TempFunctionNames->Count;i++) StoredFunctionNames->Add( TempFunctionNames->Strings[i] );
    for (i=0;i<TempFunctionDescriptions->Count;i++) StoredFunctionNames->Add( TempFunctionDescriptions->Strings[i] );
    for (i=0;i<TempFunctionReturnTypes->Count;i++) StoredFunctionNames->Add( TempFunctionReturnTypes->Strings[i] );
    for (i=0;i<TempFunctionReturnValues->Count;i++) StoredFunctionNames->Add( TempFunctionReturnValues->Strings[i] );
    for (i=0;i<TempFlagNames->Count;i++) fFlagNames->Add( TempFlagNames->Strings[i] );
    for (i=0;i<TempFlagDescriptions->Count;i++) fFlagDescriptions->Add( TempFlagDescriptions->Strings[i] );
    for (i=0;i<TempParameterNames->Count;i++) StoredParameterNames->Add( TempParameterNames->Strings[i] );
    for (i=0;i<TempParameterDescriptions->Count;i++) StoredParameterDescriptions->Add( TempParameterDescriptions->Strings[i] );
    for (i=0;i<TempParameterTypes->Count;i++) StoredParameterTypes->Add( TempParameterTypes->Strings[i] );
    for (i=0;i<TempParameterModifiers->Count;i++) StoredParameterModifiers->Add( TempParameterModifiers->Strings[i] );

*/
    TempFunctionNames->Clear();
    TempFunctionDescriptions->Clear();
    TempFunctionReturnTypes->Clear();
    TempFunctionReturnValues->Clear();
    TempParameterNames->Clear();
    TempParameterDescriptions->Clear();
    TempParameterTypes->Clear();
    TempParameterModifiers->Clear();
    TempFlagNames->Clear();
    TempFlagDescriptions->Clear();

    delete TempFlagDescriptions;
    delete TempFlagNames;
    delete TempParameterModifiers;
    delete TempParameterTypes;
    delete TempParameterDescriptions;
    delete TempParameterNames;
    delete TempFunctionReturnValues;
    delete TempFunctionReturnTypes;
    delete TempFunctionDescriptions;
    delete TempFunctionNames;
    }
}
catch (...) {} // end of try..catch                            

Dbg( IntToStr(StoredInterfaceNames->Count) + " Interfaces were imported SUCCESSFULLY!" );

// clean up temporary memory usage (unless you eventually want it to crash)
delete fFlagDescriptions;
delete fFlagNames;
delete StoredParameterDescriptions;
delete StoredParameterModifiers;
delete StoredParameterNames;
delete StoredParameterTypes;
delete StoredFunctionReturnValues;
delete StoredFunctionReturnTypes;
delete StoredFunctionDescriptions;
delete StoredFunctionNames;
delete StoredInterfaceDescriptions;
delete StoredInterfaceNames;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//brians code ends here Hi Hi Hi Hi Hi Hi Hi Hi Hi Hi Hey Hey Hey STOP NOW !!
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
fImportStrings = new TStringList();
fImportRow = 0;
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
delete fImportStrings;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//brians code back up further Hi Hi Hi Hi Hi Hi Hi Hi Hi Hi Hey Hey Hey STOP NOW !!
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall TMainForm::ImportFile( AnsiString pFileName )
{
DebuggingForm->Caption = "Importing DirectX data?Files.... debugging output";
DebuggingForm->ListBox->Items->Clear();

// load the file into a temporary TStringList
ImportStrings->Clear();

// then copy it to another TStringList so that there are no blank rows
TStringList* temp_ImportStrings = new TStringList();
temp_ImportStrings->LoadFromFile( pFileName );
for (int i=0;i<temp_ImportStrings->Count;i++) if (!temp_ImportStrings->Strings[i].TrimLeft().TrimRight().IsEmpty()) ImportStrings->Add(temp_ImportStrings->Strings[i].TrimLeft().TrimRight());
delete temp_ImportStrings;

Dbg( "Importing file: '" + pFileName + "'.... " );
ImportRow = 0;

// scan through each of the 'strings' in the input file & parse as we go.

// check the file to make sure it contains structures
if (IsKeyword(ImportString,"Structures"))
    {
    Dbg( "... '" + pFileName + "' is a 'Structures' definition file.... attempting to parse...");
    ImportRow++;
    ImportStructures( pFileName );
    }
else if (IsKeyword(ImportString,"Interfaces"))
    {
    Dbg( "... '" + pFileName + "' is an 'Interfaces' definition file.... attempting to parse...");
    ImportRow++;
    ImportInterfaces( pFileName );
    }
else
    ERROR( "There was no file type definition at the top of the file..." );
}
//---------------------------------------------------------------------------
void        __fastcall  TMainForm::FSetImportRow( int row )
{
if (row<fImportStrings->Count) Dbg( "[Row #" + IntToStr(row) + ": '" + fImportStrings->Strings[row] + "']" );
fImportRow = row;
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::FGetImportString()
{
if (fImportRow<fImportStrings->Count) return fImportStrings->Strings[fImportRow];
return "";
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::RemoveHungarianNotation( AnsiString hungarian_string )
{
//Dbg("<><><><><><><><><><><><><><><><><><><><><><><> Inside RemoveHungarianNotation() <><><><><><><><><><><><><><><><><><><><><><><><>");
//Dbg("hungarian_string = '" + hungarian_string + "'");
char* p = hungarian_string.c_str();
int first_uppercase_char_location = 0;
for (int i=1;i<hungarian_string.Length();i++)
    {
    if ((*p >= 'A') && (*p <= 'Z'))
        {
        first_uppercase_char_location = i;
        break;
        }
    p++;
    }
// lprDest
// rDes
//Dbg("First uppercase character is at pos #" + IntToStr(first_uppercase_char_location));
AnsiString resulting_string = hungarian_string.SubString( first_uppercase_char_location, hungarian_string.Length()-first_uppercase_char_location+1 );
//Dbg("resulting_string = '" + resulting_string + "'" );
//Dbg("<><><><><><><><><><><><><><><><><><><><><><><> Inside RemoveHungarianNotation() <><><><><><><><><><><><><><><><><><><><><><><><>");
return resulting_string;
}
//---------------------------------------------------------------------------
AnsiString  __fastcall  TMainForm::EncodeValue( AnsiString code, int value )
{
// build an encoded string
return "::[" + code + "," + IntToStr(value) + "]::";
}
//---------------------------------------------------------------------------
int         __fastcall  TMainForm::DecodeValue( AnsiString code, AnsiString encoded_string, AnsiString* decoded_string )
{
//Dbg( ">>> Inside DecodeValue().. Code is: '" + code + "'" );
//Dbg( ">>> Inside DecodeValue().. Encoded string is: '" + encoded_string + "'" );
while (true)
    {
    // determine if this supplied string has been encoded
    int start_pos = encoded_string.Pos("::[");
    int end_pos = encoded_string.Pos("]::");
    if ((start_pos==0) || (end_pos==0))
        {
//        if (start_pos==0) Dbg( "could not find a start token" );
//        if (end_pos==0) Dbg( "could not find an end token" );
        *decoded_string = encoded_string;
        return -1;    // exit if not encoded
        }

    // if (a) the current encoded_section contains the valid code
    // (b) a comma is inside the encoded string
//    Dbg("encoded section has been found. extracting");
    AnsiString encoded_section = encoded_string.SubString( start_pos+3, end_pos-start_pos-3 );
//    Dbg("encoded_section == '" + encoded_section + "'" );
//    Dbg("searching for a comma");
    int comma_pos = encoded_section.Pos(",");
    if ((encoded_section.Pos(code)==1) && (comma_pos != 0))
        // return the encoded value
        {
//        Dbg("Found relevant code name + comma in encoded section");
        AnsiString value_string = encoded_section.SubString( comma_pos+1, encoded_section.Length() - comma_pos );
//        Dbg("value_string == '" + value_string + "'" );
        encoded_string.Delete( start_pos, end_pos-start_pos+3 );
        *decoded_string = encoded_string;
        return StrToInt( value_string );
        }

    // delete the encoded section and continue to look for more encoded sections
//    Dbg("did not find appropriate encoded section... temporarily deleting subsection of string");
    encoded_string.Delete( start_pos, end_pos-start_pos+3 );
//    Dbg("after deleting subsection of string, the encoded_string == '" + encoded_string + "'" );
    }
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsFlag( AnsiString string )
{
// if there is a space, maybe its a flag with a description on the same line
int temp_pos = string.Pos(" ");
if (temp_pos!=0) string = string.SubString(1,temp_pos-1);
//Dbg( "Checking FLAG RULES... string == '" + string + "'" );
char* p = string.c_str();
bool all_numeric = true;
for (int i=1;i<string.Length();i++)
    {
    if (!( ((*p >= '0') && (*p <= '9')) ||
            (*p == '_') ))
        {
        all_numeric = false;
        break;
        }
    p++;
    }
//if (all_numeric) Dbg("Flag is ALL numeric, IS a flag");
//if (!all_numeric) Dbg("Flag is not ALL numeric, therefore is NOT a flag");
if (all_numeric) return false;
bool all_uppercase = true;
for (int i=1;i<string.Length();i++)
    {
    if (!( ((*p >= 'A') && (*p <= 'Z')) ||
           ((*p >= '0') && (*p <= '9')) ||
            (*p == '_') ))
        {
        all_uppercase = false;
        break;
        }
    p++;
    }
//if (all_uppercase) Dbg("Flag is ALL uppercase, IS a flag");
//if (!all_uppercase) Dbg("Flag is not ALL uppercase, therefore is NOT a flag");
return (all_uppercase);
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsThisAndThat( TStringList* known_items, AnsiString string, AnsiString* this_string, AnsiString* that_string )
{
int and_pos = string.Pos(" and ");
if (and_pos==0) return false;
*this_string = string.SubString(1,and_pos-1);
//Dbg("inside IsThisAndThat():: this_string == '" + *this_string + "'" );
*that_string = string.SubString( and_pos+5, ImportString.Length() - and_pos+1 );
//Dbg("inside IsThisAndThat():: that_string == '" + *that_string + "'" );
return (IsMemberName(known_items,*this_string) && IsMemberName(known_items,*that_string));
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsThisAndThatAndSomethingElse( TStringList* known_items, AnsiString string, TStringList* result_strings )
{
result_strings->Clear();
//Dbg("inside IsThisAndThatAndSomethingElse():: before checking for comma, string == '" + string + "'" );
int comma_pos = string.Pos(",");
while (comma_pos!=0)
    {
    result_strings->Add( string.SubString(1,comma_pos-1) );
//Dbg("inside IsThisAndThatAndSomethingElse():: found a comma... added substring == '" + string.SubString(1,comma_pos-1) + "'" );
    string.Delete(1,comma_pos+1);
//    Dbg("inside IsThisAndThatAndSomethingElse():: after deletion, string == '" + string + "'" );
//Dbg("inside IsThisAndThatAndSomethingElse():: before checking for comma, string == '" + string + "'" );
    comma_pos = string.Pos(",");
    }
//Dbg("inside IsThisAndThatAndSomethingElse():: string == '" + string + "'" );
//Dbg("inside IsThisAndThatAndSomethingElse():: checking for 'and'" );
int and_pos = string.Pos("and ");
if (and_pos!=0)
    {
//    Dbg("inside IsThisAndThatAndSomethingElse():: found an 'and'... adding substring '" + string.SubString(and_pos+4,string.Length() - (and_pos+3)) + "'" );
    result_strings->Add( string.SubString(and_pos+4,string.Length() - (and_pos+3)) );
    }
bool result = false;
for (int i=0;i<result_strings->Count;i++) result |= IsMemberName(known_items,result_strings->Strings[i]);
return result;
}
//---------------------------------------------------------------------------
void    __fastcall  TMainForm::ImportStructures(  AnsiString pFileName )
{
// Immediately following the 'Structures' keyword is a list of the different
// structures defined inside the file.
// By scanning rows for structure names we build a list of the various
// structures that we are going to parse.
// When a structure name is repeated it signifies that the first structure
// definition is following
// convert it as required
fStructureNames = new TStringList();
fStructureDescriptions = new TStringList();
fMemberNames = new TStringList();
fMemberTypes = new TStringList();
fMemberDescriptions = new TStringList();
fFlagNames = new TStringList();
fFlagDescriptions = new TStringList();

AnsiString foundation_class_name = ImportString;
int foundation_class_page = -1;
for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
    {
    if (DirectXFoundationClassPages->Pages[i]->Caption==foundation_class_name)
        {
        foundation_class_page = i;
        Dbg( "Importing '" + foundation_class_name + "' structures!!");
        break;
        }
    }
if (foundation_class_page==-1) ERROR("Did not find a valid foundation class name in the import file!");
DirectXFoundationClassPages->ActivePage = DirectXFoundationClassPages->Pages[foundation_class_page];
DirectXFoundationClassPagesChange(this);
ImportRow++;

Dbg( ">>> determining the various structures included in this file <<<");
bool found_first_repeated_structure = false;
do
    {
    // check the current row to see if it contains a unique structure name
    Dbg( "checking row for a unique structure name" );
    if (IsStructureName(fStructureNames, ImportString))
        {
        Dbg( "found a repetition of a structure name: '" + ImportString + "'... assuming the relevant structure list is finished." );
        found_first_repeated_structure = true;
        break;  // continue at end of this for loop
        }

    // if we didn't find a unique structure name
    if (!found_first_repeated_structure)
        {
        // check that the row conforms to the structure naming rules
        if (IsValidStructureName(ImportString))
            {
            Dbg( "found a valid structure name: '" + ImportString + "'... adding to internal structures list" );
            fStructureNames->Add( ImportString );
            ImportRow++;
//            if (ImportRow >= ImportStrings->Count) ERROR( "scanned all rows but did not find a structure name repeated after the 'Structures included in this file' block...");
            }
        else ERROR( "scanned all rows but did not find a structure name repeated after the 'Structures included in this file' block...");
        }
    }
while (!found_first_repeated_structure);

// since we now know how many structures we need to parse, attempt to parse that many structures
for (int structure_index=0;structure_index<fStructureNames->Count;structure_index++)
    {
//    if (ImportRow >= ImportStrings->Count) ERROR( "scanned all rows but did not find the relevant structure definition for '" + fStructureNames->Strings[structure_index] + "'" );
    Dbg( ">>> parsing structure definition #" + IntToStr(structure_index) + " <<<" );

    // clear internal storage
    TStringList* TempMemberNames = new TStringList();
    TStringList* TempMemberTypes = new TStringList();
    TStringList* TempMemberDescriptions = new TStringList();
    TStringList* TempFlagNames = new TStringList();
    TStringList* TempFlagDescriptions = new TStringList();

    // since we already know the names, check that the definition is appropriate but otherwise
    // ignore this row.
    Dbg( "Checking structure name to make sure that a definition is following");
    if (ImportString!=fStructureNames->Strings[structure_index]) ERROR( "Did not find the expected structure name: '" + fStructureNames->Strings[structure_index] + "'" );
    Dbg( "name is valid for structure: '" + fStructureNames->Strings[structure_index] + "'..." );
    Dbg( "... Assuming that a structure definition for '" + fStructureNames->Strings[structure_index] + "' is following" );
    Msg( "Parsing Structure: '" + fStructureNames->Strings[structure_index] + "'" );
    ImportRow++;

    // simply add the structure description
    Dbg( "Assuming row #" + IntToStr( ImportRow ) + " is a description for structure: '" + fStructureNames->Strings[structure_index] + "'" );
    fStructureDescriptions->Add( ImportString );
    ImportRow++;

    // by all rights we _must_ have a "struct" somewhere in this row
    Dbg( "checking row for a 'struct' construct..." );
    if (!IsStartOfStruct(ImportString)) ERROR( "There is no starting point for the 'struct' definition for the object '" + fStructureNames->Strings[structure_index] + "'" );
    Dbg( "... a 'struct' code declaration has been found..." );
    ImportRow++;

    // the following rows may occur anywhere.... hence we loop till we find the end of the 'struct' definition
    bool end_of_struct = false;
    int  inside_union_count = 0;
    Dbg( "parsing 'struct' definition..." );
    do
        {
        // skip out when we find the end-token
        if (IsEndOfStruct(ImportString))
            {
            Dbg( "Found the end of the 'struct' code declaration." );
            ImportRow++;
            end_of_struct = true;
            continue;
            }
        // ignore any 'union' declarations
        if (IsStartOfUnion(ImportString))
            {
            Dbg( "Found the start of a 'union' declaration... ignoring the 'union' row..." );
            ImportRow++;
            inside_union_count++;
            continue;
            }

        // ignore anything special while inside a 'union' declaration
        if (inside_union_count>0)
            {
            // ignore any '{' declarations
            if (ContainsStartBracket(ImportString))
                {
                Dbg( "ignoring row because it contains a '{' within a 'union' declaration" );
                ImportRow++;
                continue;
                }
            // ignore any '};' declarations  part of the 'union'
            if (IsEndOfUnion(ImportString))
                {
                Dbg( "Found the end of a 'union' declaration" );
                ImportRow++;
                inside_union_count--;
                continue;
                }
            }

        // determine the member type & name.
        
        // input:     LPRECT              lprDest;         // some comment
        // logic: LPRECT              lprDest;         // some comment
        // logic:       ^
        // result: LPRECT
        AnsiString temp_string = ImportString.TrimLeft();
        int temp_pos = temp_string.Pos(" ");
        AnsiString member_type = temp_string.SubString(1,temp_pos-1);
        Dbg( "found a structure member type: '" + member_type + "'" );
        TempMemberTypes->Add( member_type );
        // input: LPRECT              lprDest;         // some comment
        // logic: lprDest;         // some comment
        // logic:        ^
        // result: lprDest
        AnsiString member_name = temp_string.SubString(temp_pos, temp_string.Length() - temp_pos ).TrimLeft();
        temp_pos = member_name.Pos(";");
        if (temp_pos!=0) member_name = member_name.SubString( 1, temp_pos-1 );
        Dbg( "found a structure member name: '" + member_name + "'" );
        Msg( "Parsing Structure: '" + fStructureNames->Strings[structure_index] + "':: Found Member:'" + member_type + "  " + member_name + "'" );
        TempMemberNames->Add( member_name );

        ImportRow++;
//        if (ImportRow >= ImportStrings->Count) ERROR( "scanned all rows but did not find the end of the 'struct' declaration.");
        }
    while (!end_of_struct);

    // we now expect the word "Members"... if not something has gone wrong.
    Dbg( "expecting a 'Members' keyword... comparing..." );
    if (!IsKeyword(ImportString,"Members")) ERROR( "Did not find the 'Members' keyword as expected." );
    Dbg( "row contained a valid 'Members' keyword" );
    ImportRow++;

    // we can now parse the descriptions of the various members of the structure.
    // note that some members also contain descriptions of various flags.
    // note also that other members contain "something and somethingelse" which
    // we need to distinguish between.
    Dbg( "attempting to parse the descriptions for the various structure members" );
    bool checking_last_member_flags = false;
    for (int member_index=0;member_index<TempMemberNames->Count;member_index++)
        {
        Dbg( "parsing structure member #" + IntToStr(member_index) );
        // check for the premature end of the members descriptions
        if (IsKeyword(ImportString,"QuickInfo"))
            {
            // special case: if we're checking for flags for the last member, and the check failed
            // do not check for "this and that" rules.
            if (!checking_last_member_flags) ERROR( "found a 'QuickInfo' section before all member descriptions were defined." );
            Dbg( "found a 'QuickInfo' section after checking all member descriptions. Assuming member descriptions are completed" );
            break;
            }
        if (IsKeyword(ImportString,"Remarks"))
            {
            // special case: if we're checking for flags for the last member, and the check failed
            // do not check for "this and that" rules.
            if (!checking_last_member_flags) ERROR( "found a 'Remarks' section before all member descriptions were defined." );
            Dbg( "found a 'Remarks' section after checking all member descriptions. Assuming member descriptions are completed" );
            break;
            }

        // if the current line contains the name of the appropriate member, ignore the member name &
        // store the description that follows immediately after this row.
        Dbg( "expecting member name: '" + TempMemberNames->Strings[member_index] + "'... checking to make sure that row is a valid member name" );
        if (ImportString==TempMemberNames->Strings[member_index])
            {
            Dbg( "found a valid member name: '" + TempMemberNames->Strings[member_index] + "'" );
            Msg( "Parsing Structure Member: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
            ImportRow++;

            Dbg( "assuming row #" + IntToStr(ImportRow) + " as the description for member: '" + TempMemberNames->Strings[member_index] + "'" );
            TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
            ImportRow++;

            // check if there are any flags following the member description
            Dbg( "checking to see if flags are declared following the member description" );
            if (IsFlag(ImportString))
                {
                Dbg( "found a flag rule immdediately following the member description." );
                do
                    {
                    Dbg( "adding a member flag: '" + ImportString + "'" );
                    Msg( "Found Structure Member Flag: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" + "::" + ImportString );
                    TempFlagNames->Add( ImportString);
                    // check for flag descriptions declared on the same line as the flag name
                    int temp_pos = ImportString.Pos(" ");
                    if (temp_pos!=0)
                        {
                        Dbg( "the member flag row also contains the relevant flag description" );
                        AnsiString flag_description = ImportString.SubString(temp_pos,ImportString.Length()-temp_pos).TrimLeft();
                        Dbg( "adding the member flag description: '" + flag_description + "'" );
                        TempFlagDescriptions->Add( EncodeValue("FlagStructureIndex",structure_index) + EncodeValue("FlagMemberIndex",fMemberNames->Count+member_index) + flag_description);
                        }
                    else
                        // check just to make sure that the flags do not have descriptions with them...
                        if (IsFlag(ImportString))
                            {
                            Dbg( "the member flag description is assumed to be on the next row..." );
                            ImportRow++;
                            Dbg("adding a member flag description: '" + ImportString + "'" );
                            TempFlagDescriptions->Add( EncodeValue("FlagStructureIndex",structure_index) + EncodeValue("FlagMemberIndex",fMemberNames->Count+member_index) + ImportString);
                            ImportRow++;
                            }
                        else
                            ERROR( "a flag was encountered followed immediately by another flag" );
                    }
                while (IsFlag(ImportString));
                }
            else
                Dbg( "there are no flags following the member description" );
            }
        else
            // it's not a simple member.... maybe it's a case of "this and that".... or "this, that, and something else"
            {
            Dbg( "did not find a member name where expected... checking 'this and that' rules..." );
            // if there is another member name following this one....
            if (member_index<TempMemberNames->Count-1)
                {
                Dbg( "...checking 'this and that' rules");
                AnsiString first_member_name;
                AnsiString second_member_name;
                if (IsThisAndThat(TempMemberNames,ImportString,&first_member_name,&second_member_name))
                    {
                    Dbg( "the row fits the rules for a 'this and that' type description" );
                    ImportRow++;    // skip past the "this and that" row
                    Dbg( "assuming row #" + IntToStr(ImportRow) + " is the description for members: '" + first_member_name + "' and '" + second_member_name + "'" );
                    TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                    TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                    ImportRow++;
                    member_index++;
                    }
                else
                    {
                    Dbg( "the row did not fit the 'this and that' rule. Checking 'this, that, and something else' rule..." );
                    TStringList* temp_member_names = new TStringList();
                    if (IsThisAndThatAndSomethingElse(TempMemberNames,ImportString, temp_member_names ))
                        {
                        Dbg( "the row fits the rules for a 'this, and that and something else' type description" );
                        ImportRow++;    // skip past the "this, that and something else" row
                        for (int i=0;i<temp_member_names->Count;i++)
                            {
                            Dbg( "adding member description for member: '" + temp_member_names->Strings[i] + "'" );
                            TempMemberDescriptions->Add( EncodeValue("StructureIndex",structure_index) + ImportString );
                            }
                        if (temp_member_names->Count>0) member_index += temp_member_names->Count-1;
                        ImportRow++;
                        }
                    else
                        {
                        Dbg( "Found the following member names when checking 'this, that, and something else' rule" );
                        for (int i=0;i<temp_member_names->Count;i++) Dbg( "member #" + IntToStr(i) + "=='" + temp_member_names->Strings[i] + "'" );
                        temp_member_names->Clear();
                        delete temp_member_names;
                        ERROR( "Expected but did not find a particular member name & description: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
                        }
                    temp_member_names->Clear();
                    delete temp_member_names;
                    }
                }
            else
                ERROR( "Expected but did not find a particular member name & description: '" + fStructureNames->Strings[structure_index] + "::" + TempMemberNames->Strings[member_index] + "'" );
            }
        }

    // we now expect either a "Remarks" or a "QuickInfo" section
    // check just to make sure
    // if it is a 'Remarks' section, simply skip it till we find a "QuickInfo" section.

    if (ImportString=="Remarks")
        while (ImportString!="QuickInfo")
            {
            ImportRow++;
            if (ImportRow>=ImportStrings->Count) ERROR( "Found a 'Remarks' section but did not find the 'QuickInfo' section " );
            }

//    Dbg( "Checking if the keyword 'QuickInfo' exists.");
//    Dbg( "string being checked == '" + ImportString + "'" );
    if (ImportString!="QuickInfo") ERROR( "Could not find a 'QuickInfo' section" );
    ImportRow += 5;

    // yey. the structure definition has been successfully parsed.
    Dbg( ">>>>>> Definition of structure '" + fStructureNames->Strings[structure_index] + "' was completed successfully! <<<<<<" );

    //    Dbg( "Clearing structure internal representations" );
    int i;
    for (i=0;i<TempFlagDescriptions->Count;i++) fFlagDescriptions->Add( TempFlagDescriptions->Strings[i] );
    for (i=0;i<TempFlagNames->Count;i++) fFlagNames->Add( TempFlagNames->Strings[i] );
    for (i=0;i<TempMemberDescriptions->Count;i++) fMemberDescriptions->Add( TempMemberDescriptions->Strings[i] );
    for (i=0;i<TempMemberTypes->Count;i++) fMemberTypes->Add( TempMemberTypes->Strings[i] );
    for (i=0;i<TempMemberNames->Count;i++) fMemberNames->Add( TempMemberNames->Strings[i] );

    TempFlagDescriptions->Clear();
    TempFlagNames->Clear();
    TempMemberDescriptions->Clear();
    TempMemberTypes->Clear();
    TempMemberNames->Clear();

    delete TempFlagDescriptions;
    delete TempFlagNames;
    delete TempMemberDescriptions;
    delete TempMemberTypes;
    delete TempMemberNames;
    }

Dbg( IntToStr(fStructureNames->Count) + " structures were imported SUCCESSFULLY!" );
Msg( IntToStr(fStructureNames->Count) + " structures were imported SUCCESSFULLY!" );

Dbg("");
Dbg( "Structures DUMP :-" );
Msg( "Dumping structures to database" );
Dbg("");
int structure;
int member;
int flag;
Dbg( "# of structures names == " + IntToStr(fStructureNames->Count) );
Dbg( "# of structures descriptions == " + IntToStr(fStructureDescriptions->Count) );
Dbg( "# of member names == " + IntToStr(fMemberNames->Count) );
Dbg( "# of member types == " + IntToStr(fMemberTypes->Count) );
Dbg( "# of member descriptions == " + IntToStr(fMemberDescriptions->Count) );
Dbg( "# of flag names == " + IntToStr(fFlagNames->Count) );
Dbg( "# of flag descriptions == " + IntToStr(fFlagDescriptions->Count) );

TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add( " select Foundation_Class_Index from Foundation_Classes " );
TempQuery->SQL->Add( " where Foundation_Class_Name = '" + foundation_class_name + "'" );
TempQuery->Open();
TempQuery->First();
fFoundation_Class_Index = TempQuery->FieldByName("Foundation_Class_Index")->AsInteger;
TempQuery->Close();

Structures_Table->Close();       // close and re-open so that we know for sure that the table has been opened
Structures_Table->Open();
for (structure=0;structure<fStructureNames->Count;structure++)
    {
    Dbg( "Saving Structure #" + IntToStr(structure) + ": '" + fStructureNames->Strings[structure] + "'" );
    Msg( "Saving Structure #" + IntToStr(structure) + ": '" + fStructureNames->Strings[structure] + "'" );
    Structures_Table->Append();
    Structures_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    Structures_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    Structures_Table->FieldByName("Structure_Name")->AsString = fStructureNames->Strings[structure];
    Structures_Table->FieldByName("Structure_Description")->AsString = fStructureDescriptions->Strings[structure];
    Structures_Table->FieldByName("BCB_Name")->AsString = fStructureNames->Strings[structure].LowerCase();
    Structures_Table->FieldByName("BCB_Description")->AsString = fStructureDescriptions->Strings[structure];
    Structures_Table->Post();
    // retrieve & store the unique field value (created during the Post() but not yet
    Structures_Table->Refresh();
//    TempQuery->Close();
//    TempQuery->SQL->Clear();
//    TempQuery->SQL->Add( " select Structure_Index from Structures" );
//    TempQuery->SQL->Add( " where Structure_Name = '" + fStructureNames->Strings[structure] + "'" );
//    TempQuery->Open();
//    Dbg( "TempQuery:: Structure #" + IntToStr(structure) + " [" + fStructureNames->Strings[structure] + "] unique index == " + TempQuery->FieldByName("Structure_Index")->AsString );
//    TempQuery->Close();

//    Dbg( "Structures_Table:: Structure #" + IntToStr(structure) + " [" + fStructureNames->Strings[structure] + "] unique index == " + Structures_Table->FieldByName("Structure_Index")->AsString );

//    Dbg( "original structure name == '" + fStructureNames->Strings[structure] + "'" );

    Dbg( "Unique Structure Index == " + Structures_Table->FieldByName("Structure_Index")->AsString );
    fStructureNames->Strings[structure] = EncodeValue( "UniqueStructureIndex", Structures_Table->FieldByName("Structure_Index")->AsInteger) + fStructureNames->Strings[structure];
//    TempQuery->Close();
//    Dbg( "modified structure name with unique field == '" + fStructureNames->Strings[structure] + "'" );
    }

StructureMembers_Table->Close();
StructureMembers_Table->Open();
AnsiString decoded_member_description;
AnsiString crap_to_be_ignored;
for (member=0;member<fMemberNames->Count;member++)
    {
    structure = DecodeValue( "StructureIndex", fMemberDescriptions->Strings[member], &decoded_member_description );
    int unique_structure_index = DecodeValue( "UniqueStructureIndex", fStructureNames->Strings[structure], &crap_to_be_ignored );

    Dbg( "Saving Structure Member #" + IntToStr(member) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "'" );
    Msg( "Saving Structure Member #" + IntToStr(member) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "'" );

    StructureMembers_Table->Append();
    StructureMembers_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    StructureMembers_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    StructureMembers_Table->FieldByName("Structure_Index")->AsInteger = unique_structure_index;
    StructureMembers_Table->FieldByName("Member_Name")->AsString = fMemberNames->Strings[member];
    StructureMembers_Table->FieldByName("Member_Type")->AsString = fMemberTypes->Strings[member];
    StructureMembers_Table->FieldByName("Member_Description")->AsString= decoded_member_description;
    StructureMembers_Table->FieldByName("BCB_Name")->AsString = RemoveHungarianNotation( fMemberNames->Strings[member] );
    StructureMembers_Table->FieldByName("BCB_Description")->AsString= decoded_member_description;
    StructureMembers_Table->Post();
    StructureMembers_Table->Refresh();

    Dbg( "Unique Member Index == " + StructureMembers_Table->FieldByName("Member_Index")->AsString );
    fMemberNames->Strings[member] = EncodeValue( "UniqueMemberIndex", StructureMembers_Table->FieldByName("Member_Index")->AsInteger) + fMemberNames->Strings[member];
    }

StructureMemberFlags_Table->Close();
StructureMemberFlags_Table->Open();
AnsiString decoded_flag_description;
for (flag=0;flag<fFlagNames->Count;flag++)
    {
    structure = DecodeValue( "FlagStructureIndex", fFlagDescriptions->Strings[flag], &decoded_flag_description );
    int unique_structure_index = DecodeValue( "UniqueStructureIndex", fStructureNames->Strings[structure], &crap_to_be_ignored );

    member = DecodeValue( "FlagMemberIndex", fFlagDescriptions->Strings[flag], &decoded_flag_description );
    int unique_member_index = DecodeValue( "UniqueMemberIndex", fMemberNames->Strings[member], &crap_to_be_ignored );

    Dbg( "Saving flag #" + IntToStr(flag) + " - '" + fFlagNames->Strings[flag] + "'" );
    Dbg( "Of member #" + IntToStr(member) + " - '" + fMemberNames->Strings[member] + "'" );
    Dbg( "Of structure #" + IntToStr(structure) + " - '" + fStructureNames->Strings[structure] + "'" );

    Dbg( "Saving Structure Member Flag #" + IntToStr(flag) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "::" + fFlagNames->Strings[flag] + "'" );
    Msg( "Saving Structure Member Flag #" + IntToStr(flag) + ": '" + fStructureNames->Strings[structure] + "::" + fMemberNames->Strings[member] + "::" + fFlagNames->Strings[flag] + "'" );

    StructureMemberFlags_Table->Append();
    StructureMemberFlags_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    StructureMemberFlags_Table->FieldByName("Foundation_Class_Index")->AsInteger = fFoundation_Class_Index;
    StructureMemberFlags_Table->FieldByName("Structure_Index")->AsInteger = unique_structure_index;
    StructureMemberFlags_Table->FieldByName("Member_Index")->AsInteger = unique_member_index;
    StructureMemberFlags_Table->FieldByName("Flag_Name")->AsString = fFlagNames->Strings[flag];
    StructureMemberFlags_Table->FieldByName("Flag_Description")->AsString = decoded_flag_description;
    StructureMemberFlags_Table->FieldByName("BCB_Description")->AsString = decoded_flag_description;
    StructureMemberFlags_Table->Post();
    }

// clean up usage
fFlagDescriptions->Clear();
fFlagNames->Clear();
fMemberDescriptions->Clear();
fMemberTypes->Clear();
fMemberNames->Clear();
fStructureDescriptions->Clear();
fStructureNames->Clear();

Dbg( "Done. All structures were imported successfully!" );
Msg( "Done. All structures were imported successfully!" );

delete fFlagDescriptions;
delete fFlagNames;
delete fMemberDescriptions;
delete fMemberTypes;
delete fMemberNames;
delete fStructureDescriptions;
delete fStructureNames;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsStructureName( TStringList* structure_name_strings, AnsiString structure_name )
{
for (int i=0;i<structure_name_strings->Count;i++) if (structure_name_strings->Strings[i]==structure_name) return true;
return false;
}
//---------------------------------------------------------------------------
bool    __fastcall  TMainForm::IsMemberName( TStringList* member_name_strings, AnsiString member_name )
{
for (int i=0;i<member_name_strings->Count;i++) if (member_name_strings->Strings[i]==member_name) return true;
return false;
}
//---------------------------------------------------------------------------
void    __fastcall  TMainForm::UpdateStructures()
{
// build a list of current known structures & put them in the relevant DirectX & BCB combo boxes
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Structure_Name, BCB_Name from Structures " );
TempQuery->SQL->Add(" where Foundation_Class_Index = " + IntToStr(fFoundation_Class_Index) );
TempQuery->SQL->Add(" order by Structure_Name ");
TempQuery->Open();
TempQuery->First();
int orig_itemindex = StructureName_ComboBox->ItemIndex;
if (orig_itemindex==-1) orig_itemindex = 0;
StructureName_ComboBox->Items->Clear();
ObjectName_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    StructureName_ComboBox->Items->Add(TempQuery->FieldByName("Structure_Name")->AsString);
    ObjectName_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
StructureName_ComboBox->ItemIndex = orig_itemindex;
ObjectName_ComboBox->ItemIndex = orig_itemindex;

// jump to the relevant structure
if (StructureName_ComboBox->Items->Count>0)
    {
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Structure_Index from Structures ");
    TempQuery->SQL->Add(" where Structure_Name = '" + StructureName_ComboBox->Items->Strings[StructureName_ComboBox->ItemIndex] + "' " );
    TempQuery->SQL->Add(" and Foundation_Class_Index = " + IntToStr(fFoundation_Class_Index) );
    TempQuery->Open();
    fStructure_Index = TempQuery->FieldByName("Structure_Index")->AsInteger;
    TempQuery->Close();
    Structures_Table->Close();
    Structures_Table->IndexFieldNames = "Structure_Index";
    Structures_Table->Open ();
    Structures_Table->SetKey();
    Structures_Table->FieldByName("Structure_Index")->AsInteger = fStructure_Index;
    Structures_Table->GotoNearest();
    }

// build the relevant #15 of 55 label
fStructure_Count = 0;
if (StructureName_ComboBox->Items->Count>0)
    {
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select count(Structure_Name) from Structures " );
    TempQuery->SQL->Add(" where Foundation_Class_Index = " + IntToStr(fFoundation_Class_Index) );
    TempQuery->Open();
    fStructure_Count = TempQuery->Fields[0]->AsInteger;
    TempQuery->Close();
    }
// do/don't display/edit bits as required
bool structures_exist = false;
AnsiString structures_str = "No Structures";
if (fStructure_Count!=0)
    {
    structures_exist = true;
    structures_str = "Structure #" + IntToStr(StructureName_ComboBox->ItemIndex+1) + " of " + IntToStr(fStructure_Count);
    }
DD_StructureCount_Label->Caption = structures_str;
DS_StructureCount_Label->Caption = structures_str;
DI_StructureCount_Label->Caption = structures_str;
DP_StructureCount_Label->Caption = structures_str;
DSET_StructureCount_Label->Caption = structures_str;
D3IM_StructureCount_Label->Caption = structures_str;
D3RM_StructureCount_Label->Caption = structures_str;
DD_StructureEdit_Label->Visible = structures_exist;
DD_StructureName_Label->Visible = structures_exist;
DD_StructureDesc_Label->Visible = structures_exist;
DS_StructureEdit_Label->Visible = structures_exist;
DS_StructureName_Label->Visible = structures_exist;
DS_StructureDesc_Label->Visible = structures_exist;
DI_StructureEdit_Label->Visible = structures_exist;
DI_StructureName_Label->Visible = structures_exist;
DI_StructureDesc_Label->Visible = structures_exist;
DP_StructureEdit_Label->Visible = structures_exist;
DP_StructureName_Label->Visible = structures_exist;
DP_StructureDesc_Label->Visible = structures_exist;
DSET_StructureEdit_Label->Visible = structures_exist;
DSET_StructureName_Label->Visible = structures_exist;
DSET_StructureDesc_Label->Visible = structures_exist;
D3IM_StructureEdit_Label->Visible = structures_exist;
D3IM_StructureName_Label->Visible = structures_exist;
D3IM_StructureDesc_Label->Visible = structures_exist;
D3RM_StructureEdit_Label->Visible = structures_exist;
D3RM_StructureName_Label->Visible = structures_exist;
D3RM_StructureDesc_Label->Visible = structures_exist;
ResetObjectDesc_Button->Visible = structures_exist;
StructureName_ComboBox->Visible = structures_exist;
StructureName_Edit->Visible = structures_exist;
ObjectName_ComboBox->Visible = structures_exist;
ObjectName_Edit->Visible = structures_exist;
ObjectDesc_Memo->Visible = structures_exist;
ObjectStatus_RadioGroup->Visible = structures_exist;
}
//---------------------------------------------------------------------------
void    __fastcall  TMainForm::UpdateMembers()
{
// build a list of current known members & put them in the relevant DirectX & BCB combo boxes
fMember_Count = 0;
if (fStructure_Count>0)
    {
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Member_Name, BCB_Name from Structure_Members " );
    TempQuery->SQL->Add(" where Structure_Index = " + IntToStr(fStructure_Index) );
    TempQuery->SQL->Add(" order by Member_Name ");
    TempQuery->Open();
    TempQuery->First();
    int orig_itemindex = MemberName_ComboBox->ItemIndex;
    if (orig_itemindex==-1) orig_itemindex = 0;
    MemberName_ComboBox->Items->Clear();
    PropertyName_ComboBox->Items->Clear();
    while (!TempQuery->Eof)
        {
        MemberName_ComboBox->Items->Add(TempQuery->FieldByName("Member_Name")->AsString);
        PropertyName_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
        TempQuery->Next();
        }
    TempQuery->Close();
    MemberName_ComboBox->ItemIndex = orig_itemindex;
    PropertyName_ComboBox->ItemIndex = orig_itemindex;

    // jump to the relevant member
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Member_Index from Structure_Members ");
    TempQuery->SQL->Add(" where Member_Name = '" + MemberName_ComboBox->Items->Strings[MemberName_ComboBox->ItemIndex] + "' " );
    TempQuery->SQL->Add(" and Structure_Index = " + IntToStr(fStructure_Index) );
    TempQuery->Open();
    fMember_Index = TempQuery->FieldByName("Member_Index")->AsInteger;
    TempQuery->Close();
    StructureMembers_Table->Close();
    StructureMembers_Table->IndexFieldNames = "Member_Index";
    StructureMembers_Table->Open();
    StructureMembers_Table->SetKey();
    StructureMembers_Table->FieldByName("Member_Index")->AsInteger = fMember_Index;
    StructureMembers_Table->GotoNearest();

    // build the relevant #15 of 55 label
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select count(Member_Name) from Structure_Members " );
    TempQuery->SQL->Add(" where Structure_Index = " + IntToStr(fStructure_Index) );
    TempQuery->Open();
    fMember_Count = TempQuery->Fields[0]->AsInteger;
    TempQuery->Close();
    }

// do/don't display/edit bits as required
bool members_exist = false;
AnsiString members_str = "No Members";
if ((fStructure_Count!=0) &&
    (fMember_Count!=0))
    {
    members_exist = true;
    members_str = "Member #" + IntToStr(MemberName_ComboBox->ItemIndex+1) + " of " + IntToStr(fMember_Count);
    }
DD_MemberCount_Label->Caption = members_str;
DS_MemberCount_Label->Caption = members_str;
DI_MemberCount_Label->Caption = members_str;
DP_MemberCount_Label->Caption = members_str;
DSET_MemberCount_Label->Caption = members_str;
D3IM_MemberCount_Label->Caption = members_str;
D3RM_MemberCount_Label->Caption = members_str;
DD_PropertyEdit_Label->Visible = members_exist;
DD_PropertyName_Label->Visible = members_exist;
DD_PropertyDesc_Label->Visible = members_exist;
DS_PropertyEdit_Label->Visible = members_exist;
DS_PropertyName_Label->Visible = members_exist;
DS_PropertyDesc_Label->Visible = members_exist;
DI_PropertyEdit_Label->Visible = members_exist;
DI_PropertyName_Label->Visible = members_exist;
DI_PropertyDesc_Label->Visible = members_exist;
DP_PropertyEdit_Label->Visible = members_exist;
DP_PropertyName_Label->Visible = members_exist;
DP_PropertyDesc_Label->Visible = members_exist;
DSET_PropertyEdit_Label->Visible = members_exist;
DSET_PropertyName_Label->Visible = members_exist;
DSET_PropertyDesc_Label->Visible = members_exist;
D3IM_PropertyEdit_Label->Visible = members_exist;
D3IM_PropertyName_Label->Visible = members_exist;
D3IM_PropertyDesc_Label->Visible = members_exist;
D3RM_PropertyEdit_Label->Visible = members_exist;
D3RM_PropertyName_Label->Visible = members_exist;
D3RM_PropertyDesc_Label->Visible = members_exist;
ResetPropertyDesc_Button->Visible = members_exist;
MemberName_ComboBox->Visible = members_exist;
MemberName_Edit->Visible = members_exist;
PropertyName_ComboBox->Visible = members_exist;
PropertyName_Edit->Visible = members_exist;
PropertyDesc_Memo->Visible = members_exist;
PropertyStatus_RadioGroup->Visible = members_exist;
}
//---------------------------------------------------------------------------
void    __fastcall  TMainForm::UpdateFlags()
{
// build a list of current known flags & put them in the relevant DirectX & BCB combo boxes
fFlag_Count = 0;
if ((fStructure_Count!=0) &&
    (fMember_Count!=0))
    {
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Flag_Name from Structure_Member_Flags ");
    TempQuery->SQL->Add(" where Member_Index = " + IntToStr( fMember_Index ) );
    TempQuery->SQL->Add(" order by Flag_Name ");
    TempQuery->Open();
    TempQuery->First();
    int orig_itemindex = FlagName_ComboBox->ItemIndex;
    if (orig_itemindex==-1) orig_itemindex = 0;
    FlagName_ComboBox->Items->Clear();
    while (!TempQuery->Eof)
        {
        FlagName_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
        TempQuery->Next();
        }
    TempQuery->Close();
    FlagName_ComboBox->ItemIndex = orig_itemindex;

    // jump to the relevant flag
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Flag_Index from Structure_Member_Flags ");
    TempQuery->SQL->Add(" where Flag_Name = '" + FlagName_ComboBox->Items->Strings[FlagName_ComboBox->ItemIndex] + "' " );
    TempQuery->SQL->Add(" and Member_Index = " + IntToStr(fMember_Index) );
    TempQuery->Open();
    fFlag_Index = TempQuery->FieldByName("Flag_Index")->AsInteger;
    TempQuery->Close();
    StructureMemberFlags_Table->Close();
    StructureMemberFlags_Table->IndexFieldNames = "Flag_Index";
    StructureMemberFlags_Table->Open ();
    StructureMemberFlags_Table->SetKey();
    StructureMemberFlags_Table->FieldByName("Flag_Index")->AsInteger = fFlag_Index;
    StructureMemberFlags_Table->GotoNearest();

    // build the relevant #15 of 55 label
    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select count(Flag_Name) from Structure_Member_Flags " );
    TempQuery->SQL->Add(" where Member_Index = " + IntToStr(fMember_Index) );
    TempQuery->Open();
    fFlag_Count = TempQuery->Fields[0]->AsInteger;
    TempQuery->Close();
    }

// do/don't display/edit bits as required
bool flags_exist = false;
AnsiString flags_str = "No Flags";
if ((fStructure_Count!=0) &&
    (fMember_Count!=0) &&
    (fFlag_Count!=0))
    {
    flags_exist = true;
    flags_str = "Flag #" + IntToStr(FlagName_ComboBox->ItemIndex+1) + " of " + IntToStr(fFlag_Count);
    }
DD_FlagCount_Label->Caption = flags_str;
DS_FlagCount_Label->Caption = flags_str;
DI_FlagCount_Label->Caption = flags_str;
DP_FlagCount_Label->Caption = flags_str;
DSET_FlagCount_Label->Caption = flags_str;
D3IM_FlagCount_Label->Caption = flags_str;
D3RM_FlagCount_Label->Caption = flags_str;
DD_FlagName_Label->Visible = flags_exist;
DD_FlagDesc_Label->Visible = flags_exist;
DS_FlagName_Label->Visible = flags_exist;
DS_FlagDesc_Label->Visible = flags_exist;
DI_FlagName_Label->Visible = flags_exist;
DI_FlagDesc_Label->Visible = flags_exist;
DP_FlagName_Label->Visible = flags_exist;
DP_FlagDesc_Label->Visible = flags_exist;
DSET_FlagName_Label->Visible = flags_exist;
DSET_FlagDesc_Label->Visible = flags_exist;
D3IM_FlagName_Label->Visible = flags_exist;
D3IM_FlagDesc_Label->Visible = flags_exist;
D3RM_FlagName_Label->Visible = flags_exist;
D3RM_FlagDesc_Label->Visible = flags_exist;
ResetFlagDesc_Button->Visible = flags_exist;
FlagName_ComboBox->Visible = flags_exist;
FlagDesc_Memo->Visible = flags_exist;
FlagStatus_RadioGroup->Visible = flags_exist;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormActivate(TObject *Sender)
{
DirectX_To_BCB->Connected = true;
Structures_Table->Active = true;
StructureMembers_Table->Active = true;
StructureMemberFlags_Table->Active = true;
if (Structures_Table->RecordCount==0) Structures_Table->Insert(); else Structures_Table->Edit();
if (StructureMembers_Table->RecordCount==0) StructureMembers_Table->Insert(); else StructureMembers_Table->Edit();
if (StructureMemberFlags_Table->RecordCount==0) StructureMemberFlags_Table->Insert(); else StructureMemberFlags_Table->Edit();

// force the main pages to change so that the OnChange event occurs
DirectXFoundationClassPagesChange(Sender);
// force the combo boxes to contain valid information
UpdateStructures();
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDeactivate(TObject *Sender)
{
Structures_Table->Active = false;
StructureMembers_Table->Active = false;
StructureMemberFlags_Table->Active = false;
DirectX_To_BCB->Connected = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Top = 0;
Left = 0;
Width = 800;
Height = 580;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ClearAllStructuresButtonClick(TObject *Sender)
{
ShowMessage("YEEHAW! _ALL_ Structures, _ALL_ Methods, _ALL_ Flags will be cleared.");
ShowMessage("DO NOT click on the OK button until you are sure!\n\nBy terminating the application NOW without clicking on the OK button\nyou will not delete the entire database.\n\nBy clicking on the OK button below the entire database will be deleted and cannot be recovered!");
Structures_Table->Close();
Structures_Table->EmptyTable();
Structures_Table->Open();
StructureMembers_Table->Close();
StructureMembers_Table->EmptyTable();
StructureMembers_Table->Open();
StructureMemberFlags_Table->Close();
StructureMemberFlags_Table->EmptyTable();
StructureMemberFlags_Table->Open();
ShowMessage("The entire database has been deleted and can no longer be recovered!");

UpdateStructures();
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ObjectName_EditExit(TObject *Sender)
{
if (Structures_Table->Modified)
    {
    if (Structures_Table->FieldByName("Conversion_Status")->AsInteger==0) Structures_Table->FieldByName("Conversion_Status")->AsInteger = 1;
    Structures_Table->Post();
    UpdateStructures();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PropertyName_EditExit(
    TObject *Sender)
{
if (StructureMembers_Table->Modified)
    {
    if (StructureMembers_Table->FieldByName("Conversion_Status")->AsInteger==0) StructureMembers_Table->FieldByName("Conversion_Status")->AsInteger = 1;
    StructureMembers_Table->Post();
    UpdateMembers();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FlagDesc_MemoExit(
    TObject *Sender)
{
if (StructureMemberFlags_Table->Modified)
    {
    if (StructureMemberFlags_Table->FieldByName("Conversion_Status")->AsInteger==0) StructureMemberFlags_Table->FieldByName("Conversion_Status")->AsInteger = 1;
    StructureMemberFlags_Table->Post();
    UpdateFlags();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DBRadioGroup1Exit(TObject *Sender)
{
if (Structures_Table->Modified) Structures_Table->Post();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ToggleDebugButtonClick(TObject *Sender)
{
DebuggingForm->Visible = !DebuggingForm->Visible;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Label7Click(TObject *Sender)
{
ObjectName_ComboBox->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label6Click(TObject *Sender)
{
ObjectName_Edit->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label5Click(TObject *Sender)
{
//ObjectDesc_Memo->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label9Click(TObject *Sender)
{
PropertyName_ComboBox->SetFocus();    
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label14Click(TObject *Sender)
{
PropertyName_Edit->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label15Click(TObject *Sender)
{
//PropertyDesc_Memo->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label16Click(TObject *Sender)
{
//BCB_PropertyFlags_ComboBox->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Label17Click(TObject *Sender)
{
//FlagDesc_Memo->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ImportStructuresButtonClick(TObject *Sender)
{
//ShowMessage("Note: There is nothing in the import files to tell me to import DirectDraw/etc. specifically.\n\nTo import a file correctly, make sure you are on the correct DirectDraw/etc. tab\nbefore selecting the relevant input file.");
if (ImportFileDialog->Execute()) ImportFile( ImportFileDialog->FileName );
UpdateStructures();
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DirectXFoundationClassPagesChange(TObject *Sender)
{
int active_page = -1;
for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
    if (DirectXFoundationClassPages->Pages[i]->Caption==DirectXFoundationClassPages->ActivePage->Caption)
        {
        active_page = i;
        break;
        }
switch (active_page)
    {
    case 0: // DirectDraw
            DebugLabel = DD_DebugLabel; break;
    case 1: // DirectSound
            DebugLabel = DS_DebugLabel; break;
    case 2: // DirectInput
            DebugLabel = DI_DebugLabel; break;
    case 3: // DirectPlay
            DebugLabel = DP_DebugLabel; break;
    case 4: // DirectSetup
            DebugLabel = DSET_DebugLabel; break;
    case 5: // Direct3D Immediate Mode
            DebugLabel = D3IM_DebugLabel; break;
    case 6: // Direct3D Retained Mode
            DebugLabel = D3RM_DebugLabel; break;
    default: DebugLabel = DD_DebugLabel; break;
    }
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add( " select Foundation_Class_Index from Foundation_Classes " );
TempQuery->SQL->Add( " where Foundation_Class_Name = '" + DirectXFoundationClassPages->ActivePage->Caption + "'" );
TempQuery->Open();
TempQuery->First();
fFoundation_Class_Index = TempQuery->FieldByName("Foundation_Class_Index")->AsInteger;
TempQuery->Close();
UpdateStructures();
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StructureName_ComboBoxChange(TObject *Sender)
{
UpdateStructures();
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MemberName_ComboBoxChange(
    TObject *Sender)
{
UpdateMembers();
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FlagName_ComboBoxChange(
    TObject *Sender)
{
UpdateFlags();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::NextStructureButtonClick(TObject *Sender)
{
// if we're not at the last flag
if (FlagName_ComboBox->ItemIndex<FlagName_ComboBox->Items->Count-1)
    {
    // scroll to the next flag
//    ShowMessage("jumping to next flag");
    FlagName_ComboBox->ItemIndex++;
    UpdateFlags();
    }
else
    // otherwise, if we're not at the last structure member
    if (MemberName_ComboBox->ItemIndex<MemberName_ComboBox->Items->Count-1)
        {
        // go to the next structure member
//        ShowMessage("jumping to next member" );
        MemberName_ComboBox->ItemIndex++;
        UpdateMembers();
        // go to the first flag of the next structure member
//        ShowMessage("jumping to first flag" );
        FlagName_ComboBox->ItemIndex = 0;
        UpdateFlags();
        }
    else
        // otherwise, if we're not at the last structure
        if (StructureName_ComboBox->ItemIndex<StructureName_ComboBox->Items->Count-1)
            {
            // go to the next structure
//            ShowMessage("jumping to next structure");
            StructureName_ComboBox->ItemIndex++;
            UpdateStructures();
            // go to the first member of the next structure
//            ShowMessage("jumping to first member");
            MemberName_ComboBox->ItemIndex = 0;
            UpdateMembers();
            // go to the first flag of the first structure member
//            ShowMessage("jumping to first flag");
            FlagName_ComboBox->ItemIndex = 0;
            UpdateFlags();
            }
        else
            // we're at the last flag of the last member of the last structure
            {
            // go to the first structure
//            ShowMessage("jumping to last structure");
            StructureName_ComboBox->ItemIndex = 0;
            UpdateStructures();
            // go to the first member of the first structure
//            ShowMessage("jumping to last member");
            MemberName_ComboBox->ItemIndex = 0;
            UpdateMembers();
            // go to the first flag of the first structure member
//            ShowMessage("jumping to last flag");
            FlagName_ComboBox->ItemIndex = 0;
            UpdateFlags();
            }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PrevStructureButtonClick(TObject *Sender)
{
// if we're not at the first flag
if (FlagName_ComboBox->ItemIndex>0)
    {
//    ShowMessage("jumping to previous flag");
    // scroll to the previous flag
    FlagName_ComboBox->ItemIndex--;
    UpdateFlags();
    }
else
    // otherwise, if we're not at the first structure member
    if (MemberName_ComboBox->ItemIndex>0)
        {
//        ShowMessage("jumping to previous member/last flag");
        // go to the previous structure member
        MemberName_ComboBox->ItemIndex--;
        UpdateMembers();
        // go to the last flag of the previous structure member
        UpdateFlags();
        FlagName_ComboBox->ItemIndex = FlagName_ComboBox->Items->Count-1;
        UpdateFlags();
        }
    else
        // otherwise, if we're not at the first structure
        if (StructureName_ComboBox->ItemIndex>0)
            {
//            ShowMessage("jumping to previous structure / last member / last flag");
            // go to the previous structure
            StructureName_ComboBox->ItemIndex--;
            UpdateStructures();
            // go to the last member of the previous structure
            UpdateMembers();
            MemberName_ComboBox->ItemIndex = MemberName_ComboBox->Items->Count-1;
            UpdateMembers();
            // go to the last flag of the last structure member
            UpdateFlags();
            FlagName_ComboBox->ItemIndex = FlagName_ComboBox->Items->Count-1;
            UpdateFlags();
            }
        else
            // we're at the first flag of the first member of the first structure
            {
//            ShowMessage("jumping to last structure / last member / last flag" );
            // go to the last structure
            StructureName_ComboBox->ItemIndex = StructureName_ComboBox->Items->Count-1;
            UpdateStructures();
            // go to the last member of the last structure
            UpdateMembers();
            MemberName_ComboBox->ItemIndex = MemberName_ComboBox->Items->Count-1;
            UpdateMembers();
            // go to the last flag of the last structure member
            UpdateFlags();
            FlagName_ComboBox->ItemIndex = FlagName_ComboBox->Items->Count-1;
            UpdateFlags();
            }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DeleteStructureButtonClick(TObject *Sender)
{
ShowMessage("Not Yet Implemented");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ObjectName_ComboBoxChange(TObject *Sender)
{
StructureName_ComboBox->ItemIndex = ObjectName_ComboBox->ItemIndex;
StructureName_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PropertyName_ComboBoxChange(TObject *Sender)
{
MemberName_ComboBox->ItemIndex = PropertyName_ComboBox->ItemIndex;
MemberName_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton1Click(TObject *Sender)
{
if (BookmarkButton1->Caption!="Bookmark &1")
    {
    for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
        if (DirectXFoundationClassPages->ActivePage->Caption == fBookmarkPages[0])
            {
            DirectXFoundationClassPages->ActivePage = DirectXFoundationClassPages->Pages[i];
            break;
            }
    DirectXFoundationClassPagesChange(Sender);
    UpdateStructures();
    StructureName_ComboBox->ItemIndex = fBookmarkInfo[0];
    UpdateStructures();
    UpdateMembers();
    MemberName_ComboBox->ItemIndex = fBookmarkInfo[1];
    UpdateMembers();
    UpdateFlags();
    FlagName_ComboBox->ItemIndex = fBookmarkInfo[2];
    UpdateFlags();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton2Click(TObject *Sender)
{
if (BookmarkButton2->Caption!="Bookmark &2")
    {
    for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
        if (DirectXFoundationClassPages->ActivePage->Caption == fBookmarkPages[1])
            {
            DirectXFoundationClassPages->ActivePage = DirectXFoundationClassPages->Pages[i];
            break;
            }
    DirectXFoundationClassPagesChange(Sender);
    UpdateStructures();
    StructureName_ComboBox->ItemIndex = fBookmarkInfo[3];
    UpdateStructures();
    UpdateMembers();
    MemberName_ComboBox->ItemIndex = fBookmarkInfo[4];
    UpdateMembers();
    UpdateFlags();
    FlagName_ComboBox->ItemIndex = fBookmarkInfo[5];
    UpdateFlags();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton3Click(TObject *Sender)
{
if (BookmarkButton3->Caption!="Bookmark &3")
    {
    for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
        if (DirectXFoundationClassPages->ActivePage->Caption == fBookmarkPages[2])
            {
            DirectXFoundationClassPages->ActivePage = DirectXFoundationClassPages->Pages[i];
            break;
            }
    DirectXFoundationClassPagesChange(Sender);
    UpdateStructures();
    StructureName_ComboBox->ItemIndex = fBookmarkInfo[6];
    UpdateStructures();
    UpdateMembers();
    MemberName_ComboBox->ItemIndex = fBookmarkInfo[7];
    UpdateMembers();
    UpdateFlags();
    FlagName_ComboBox->ItemIndex = fBookmarkInfo[8];
    UpdateFlags();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton4Click(TObject *Sender)
{
if (BookmarkButton4->Caption!="Bookmark &4")
    {
    for (int i=0;i<DirectXFoundationClassPages->PageCount;i++)
        if (DirectXFoundationClassPages->ActivePage->Caption == fBookmarkPages[3])
            {
            DirectXFoundationClassPages->ActivePage = DirectXFoundationClassPages->Pages[i];
            break;
            }
    DirectXFoundationClassPagesChange(Sender);
    UpdateStructures();
    StructureName_ComboBox->ItemIndex = fBookmarkInfo[9];
    UpdateStructures();
    UpdateMembers();
    MemberName_ComboBox->ItemIndex = fBookmarkInfo[10];
    UpdateMembers();
    UpdateFlags();
    FlagName_ComboBox->ItemIndex = fBookmarkInfo[11];
    UpdateFlags();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton1KeyDown(TObject *Sender, WORD &Key,
    TShiftState Shift)
{
if (Shift.Contains(ssAlt))
    {
    fBookmarkPages[0] = DirectXFoundationClassPages->ActivePage->Caption;
    fBookmarkInfo[0] = StructureName_ComboBox->ItemIndex;
    fBookmarkInfo[1] = MemberName_ComboBox->ItemIndex;
    fBookmarkInfo[2] = FlagName_ComboBox->ItemIndex;
    BookmarkButton1->Caption = "&1 " + StructureName_ComboBox->Items->Strings[StructureName_ComboBox->ItemIndex] + ":" +
                        MemberName_ComboBox->Items->Strings[MemberName_ComboBox->ItemIndex] + ":" +
                        FlagName_ComboBox->Items->Strings[FlagName_ComboBox->ItemIndex];
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton2KeyDown(TObject *Sender, WORD &Key,
    TShiftState Shift)
{
if (Shift.Contains(ssAlt))
    {
    fBookmarkPages[1] = DirectXFoundationClassPages->ActivePage->Caption;
    fBookmarkInfo[3] = StructureName_ComboBox->ItemIndex;
    fBookmarkInfo[4] = MemberName_ComboBox->ItemIndex;
    fBookmarkInfo[5] = FlagName_ComboBox->ItemIndex;
    BookmarkButton2->Caption = "&2 " + StructureName_ComboBox->Items->Strings[StructureName_ComboBox->ItemIndex] + ":" +
                        MemberName_ComboBox->Items->Strings[MemberName_ComboBox->ItemIndex] + ":" +
                        FlagName_ComboBox->Items->Strings[FlagName_ComboBox->ItemIndex];
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton3KeyDown(TObject *Sender, WORD &Key,
    TShiftState Shift)
{
if (Shift.Contains(ssAlt))
    {
    fBookmarkPages[2] = DirectXFoundationClassPages->ActivePage->Caption;
    fBookmarkInfo[6] = StructureName_ComboBox->ItemIndex;
    fBookmarkInfo[7] = MemberName_ComboBox->ItemIndex;
    fBookmarkInfo[8] = FlagName_ComboBox->ItemIndex;
    BookmarkButton3->Caption = "&3 " + StructureName_ComboBox->Items->Strings[StructureName_ComboBox->ItemIndex] + ":" +
                        MemberName_ComboBox->Items->Strings[MemberName_ComboBox->ItemIndex] + ":" +
                        FlagName_ComboBox->Items->Strings[FlagName_ComboBox->ItemIndex];
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BookmarkButton4KeyDown(TObject *Sender, WORD &Key,
    TShiftState Shift)
{
if (Shift.Contains(ssAlt))
    {
    fBookmarkPages[3] = DirectXFoundationClassPages->ActivePage->Caption;
    fBookmarkInfo[9] = StructureName_ComboBox->ItemIndex;
    fBookmarkInfo[10] = MemberName_ComboBox->ItemIndex;
    fBookmarkInfo[11] = FlagName_ComboBox->ItemIndex;
    BookmarkButton4->Caption = "&4 " + StructureName_ComboBox->Items->Strings[StructureName_ComboBox->ItemIndex] + ":" +
                        MemberName_ComboBox->Items->Strings[MemberName_ComboBox->ItemIndex] + ":" +
                        FlagName_ComboBox->Items->Strings[FlagName_ComboBox->ItemIndex];
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ResetObjectDesc_ButtonClick(TObject *Sender)
{
Structures_Table->Edit();
Structures_Table->FieldByName("BCB_Description")->AsString = Structures_Table->FieldByName("Structure_Description")->AsString;
Structures_Table->Post();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ResetPropertyDesc_ButtonClick(TObject *Sender)
{
StructureMembers_Table->Edit();
StructureMembers_Table->FieldByName("BCB_Description")->AsString = StructureMembers_Table->FieldByName("Member_Description")->AsString;
StructureMembers_Table->Post();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ResetFlagDesc_ButtonClick(
    TObject *Sender)
{
StructureMemberFlags_Table->Edit();
StructureMemberFlags_Table->FieldByName("BCB_Description")->AsString = StructureMemberFlags_Table->FieldByName("Flag_Description")->AsString;
StructureMemberFlags_Table->Post();
}
//---------------------------------------------------------------------------
